//! Authentication  related struct definitions

use crate::enc::Random;
use ::zeroize::Zeroize;

/// Anonymous user id
pub const USERID_ANONYMOUS: UserID = UserID([0; UserID::len()]);
/// User identifier. 16 bytes for easy uuid conversion
#[derive(Debug, Copy, Clone, PartialEq)]
pub struct UserID(pub [u8; 16]);

impl From<[u8; 16]> for UserID {
    fn from(raw: [u8; 16]) -> Self {
        UserID(raw)
    }
}

impl UserID {
    /// New random user id
    pub fn new(rand: &Random) -> Self {
        use ::core::mem::MaybeUninit;
        let mut out: MaybeUninit<[u8; 16]> = MaybeUninit::uninit();
        #[allow(unsafe_code)]
        unsafe {
            let _ = rand.fill(out.assume_init_mut());
            Self(out.assume_init())
        }
    }
    /// Anonymous user id
    pub const fn new_anonymous() -> Self {
        USERID_ANONYMOUS
    }
    /// length of the User ID in bytes
    pub const fn len() -> usize {
        16
    }
}
/// Authentication Token, basically just 32 random bytes
#[derive(Clone, Zeroize, PartialEq)]
#[zeroize(drop)]
pub struct Token(pub [u8; 32]);

impl Token {
    /// New random token, anonymous should not check this anyway
    pub fn new_anonymous(rand: &Random) -> Self {
        use ::core::mem::MaybeUninit;
        let mut out: MaybeUninit<[u8; 32]> = MaybeUninit::uninit();
        #[allow(unsafe_code)]
        unsafe {
            let _ = rand.fill(out.assume_init_mut());
            Self(out.assume_init())
        }
    }
    /// length of the token in bytes
    pub const fn len() -> usize {
        32
    }
}

impl From<[u8; 32]> for Token {
    fn from(raw: [u8; 32]) -> Self {
        Token(raw)
    }
}

// Fake debug implementation to avoid leaking tokens
impl ::core::fmt::Debug for Token {
    fn fmt(
        &self,
        f: &mut core::fmt::Formatter<'_>,
    ) -> Result<(), ::std::fmt::Error> {
        ::core::fmt::Debug::fmt("[hidden token]", f)
    }
}

/// Type of the function used to check the validity of the tokens
/// Reimplement this to use whatever database you want
pub type TokenChecker =
    fn(
        user: UserID,
        token: Token,
        service_id: ServiceID,
        domain: Domain,
    ) -> ::futures::future::BoxFuture<'static, Result<bool, ()>>;

/// domain representation
/// Security notice: internal representation is utf8, but we will
/// further limit to a "safe" subset of utf8
// SECURITY: TODO: limit to a subset of utf8
#[derive(Debug, Clone, PartialEq)]
pub struct Domain(pub String);

impl TryFrom<&[u8]> for Domain {
    type Error = ();
    fn try_from(raw: &[u8]) -> Result<Self, ()> {
        let domain_string = match String::from_utf8(raw.into()) {
            Ok(domain_string) => domain_string,
            Err(_) => return Err(()),
        };
        Ok(Domain(domain_string))
    }
}
impl From<String> for Domain {
    fn from(raw: String) -> Self {
        Self(raw)
    }
}
impl From<&str> for Domain {
    fn from(raw: &str) -> Self {
        Self(raw.to_owned())
    }
}

impl Domain {
    /// length of the User ID in bytes
    pub fn len(&self) -> usize {
        self.0.len()
    }
}

/// Reserve the first service ID for the authentication service
pub const SERVICEID_AUTH: ServiceID = ServiceID([0; 16]);
/// The Service ID is a UUID associated with the service.
#[derive(Debug, Copy, Clone, PartialEq)]
pub struct ServiceID(pub [u8; 16]);

impl From<[u8; 16]> for ServiceID {
    fn from(raw: [u8; 16]) -> Self {
        ServiceID(raw)
    }
}

impl ServiceID {
    /// length of the User ID in bytes
    pub const fn len() -> usize {
        16
    }
    /// read the service id as bytes
    pub fn as_bytes(&self) -> &[u8; 16] {
        &self.0
    }
}
