//! Handhsake handling

pub mod dirsync;
#[cfg(test)]
mod tests;
pub(crate) mod tracker;

use crate::{
    connection::ProtocolVersion,
    enc::sym::{HeadLen, TagLen},
};
use ::num_traits::FromPrimitive;

/// Handshake errors
#[derive(::thiserror::Error, Debug, Copy, Clone)]
#[non_exhaustive]
pub enum Error {
    /// Error while parsing the handshake packet
    /// TODO: more detailed parsing errors
    #[error("not an handshake packet")]
    Parsing,
    /// No such Key ID
    #[error("unknown key id")]
    UnknownKeyID,
    /// Key error
    #[error("key: {0:?}")]
    Key(#[from] crate::enc::Error),
    /// Not enough data
    #[error("not enough data")]
    NotEnoughData,
    /// Could not find common cryptography
    #[error("Negotiation of keys/hkdfs/ciphers failed")]
    Negotiation,
    /// Could not generate Keys
    #[error("Key generation failed")]
    KeyGeneration,
    /// Too many client handshakes currently running
    #[error("Too many client handshakes")]
    TooManyClientHandshakes,
    /// generic internal error
    #[error("Internal tracking error")]
    InternalTracking,
    /// Handshake Timeout
    #[error("Handshake timeout")]
    Timeout,
}

/// List of possible handshakes
#[derive(
    ::num_derive::FromPrimitive,
    Debug,
    Clone,
    Copy,
    PartialEq,
    ::strum_macros::EnumString,
    ::strum_macros::IntoStaticStr,
)]
#[repr(u8)]
pub enum HandshakeID {
    /// 1-RTT Directory synchronized handshake. Fast, no forward secrecy
    #[strum(serialize = "directory_synchronized")]
    DirectorySynchronized = 0,
    /// 2-RTT Stateful exchange. Little DDos protection
    #[strum(serialize = "stateful")]
    Stateful,
    /// 3-RTT stateless exchange. Forward secrecy and ddos protection
    #[strum(serialize = "stateless")]
    Stateless,
}
impl HandshakeID {
    /// The length of the serialized field
    pub const fn len() -> usize {
        1
    }
}
/// Parsed handshake
#[derive(Debug, Clone, PartialEq)]
pub enum HandshakeData {
    /// Directory synchronized handhsake
    DirSync(dirsync::DirSync),
}

impl HandshakeData {
    /// actual length of the handshake data
    pub fn len(&self, head_len: HeadLen, tag_len: TagLen) -> usize {
        match self {
            HandshakeData::DirSync(d) => d.len(head_len, tag_len),
        }
    }
    /// Serialize into raw bytes
    /// NOTE: assumes that there is exactly asa much buffer as needed
    pub fn serialize(
        &self,
        head_len: HeadLen,
        tag_len: TagLen,
        out: &mut [u8],
    ) {
        match self {
            HandshakeData::DirSync(d) => d.serialize(head_len, tag_len, out),
        }
    }
}

/// Kind of handshake
#[derive(::num_derive::FromPrimitive, Debug, Clone, Copy)]
#[repr(u8)]
pub enum HandshakeKind {
    /// 1-RTT, Directory synchronized handshake
    /// Request
    DirSyncReq = 0,
    /// 1-RTT, Directory synchronized handshake
    /// Response
    DirSyncResp,
    /*
    /// 2-RTT, Stateful connection
    Stateful,
    ...
    /// 3 RTT, Stateless connection
    Stateless,
    ....
    */
}
impl HandshakeKind {
    /// Length of the serialized field
    pub const fn len() -> usize {
        1
    }
}

/// Parsed handshake
#[derive(Debug, Clone, PartialEq)]
pub struct Handshake {
    /// Fenrir Protocol version
    pub fenrir_version: ProtocolVersion,
    /// enum for the parsed data
    pub data: HandshakeData,
}

impl Handshake {
    /// Build new handshake from the data
    pub fn new(data: HandshakeData) -> Self {
        Handshake {
            fenrir_version: ProtocolVersion::V0,
            data,
        }
    }
    /// return the total length of the handshake
    pub fn len(&self, head_len: HeadLen, tag_len: TagLen) -> usize {
        ProtocolVersion::len()
            + HandshakeKind::len()
            + self.data.len(head_len, tag_len)
    }
    const MIN_PKT_LEN: usize = 8;
    /// Parse the packet and return the parsed handshake
    pub fn deserialize(raw: &[u8]) -> Result<Self, Error> {
        if raw.len() < Self::MIN_PKT_LEN {
            return Err(Error::NotEnoughData);
        }
        let fenrir_version = match ProtocolVersion::from_u8(raw[0]) {
            Some(fenrir_version) => fenrir_version,
            None => return Err(Error::Parsing),
        };
        let handshake_kind = match HandshakeKind::from_u8(raw[1]) {
            Some(handshake_kind) => handshake_kind,
            None => return Err(Error::Parsing),
        };
        let data = match handshake_kind {
            HandshakeKind::DirSyncReq => dirsync::Req::deserialize(&raw[2..])?,
            HandshakeKind::DirSyncResp => {
                dirsync::Resp::deserialize(&raw[2..])?
            }
        };
        Ok(Self {
            fenrir_version,
            data,
        })
    }
    /// serialize the handshake into bytes
    /// NOTE: assumes that there is exactly as much buffer as needed
    pub fn serialize(
        &self,
        head_len: HeadLen,
        tag_len: TagLen,
        out: &mut [u8],
    ) {
        out[0] = self.fenrir_version as u8;
        out[1] = match &self.data {
            HandshakeData::DirSync(d) => match d {
                dirsync::DirSync::Req(_) => HandshakeKind::DirSyncReq,
                dirsync::DirSync::Resp(_) => HandshakeKind::DirSyncResp,
            },
        } as u8;
        self.data.serialize(head_len, tag_len, &mut out[2..]);
    }
}

trait HandshakeParsing {
    fn deserialize(raw: &[u8]) -> Result<HandshakeData, Error>;
}
