//! Socket related types and functions

use ::std::net::SocketAddr;
use ::tokio::{net::UdpSocket, task::JoinHandle};

/// Pair to easily track the socket and its async listening handle
pub type SocketTracker = (SocketAddr, JoinHandle<::std::io::Result<()>>);

/// Strong typedef for a client socket address
#[derive(Debug, Copy, Clone)]
pub(crate) struct UdpClient(pub SocketAddr);
/// Strong typedef for a server socket address
#[derive(Debug, Copy, Clone)]
pub(crate) struct UdpServer(pub SocketAddr);

/// Enable some common socket options. This is just the unsafe part
fn enable_sock_opt(
    fd: ::std::os::fd::RawFd,
    option: ::libc::c_int,
    value: ::libc::c_int,
) -> ::std::io::Result<()> {
    #[allow(unsafe_code)]
    unsafe {
        #[allow(trivial_casts)]
        let val = &value as *const _ as *const ::libc::c_void;
        let size = ::core::mem::size_of_val(&value) as ::libc::socklen_t;
        // always clear the error bit before doing a new syscall
        let _ = ::std::io::Error::last_os_error();
        let ret = ::libc::setsockopt(fd, ::libc::SOL_SOCKET, option, val, size);
        if ret != 0 {
            return Err(::std::io::Error::last_os_error());
        }
    }
    Ok(())
}
/// Add an async udp listener
pub async fn bind_udp(addr: SocketAddr) -> ::std::io::Result<UdpSocket> {
    // I know, kind of a mess. but I really wanted SO_REUSE{ADDR,PORT} and
    // no-fragmenting stuff.
    // I also did not want to load another library for this.
    // feel free to simplify,
    // especially if we can avoid libc and other libraries
    // we currently use libc because it's a dependency of many other deps

    let fd: ::std::os::fd::RawFd = {
        let domain = if addr.is_ipv6() {
            ::libc::AF_INET6
        } else {
            ::libc::AF_INET
        };
        #[allow(unsafe_code)]
        let tmp = unsafe { ::libc::socket(domain, ::libc::SOCK_DGRAM, 0) };
        let lasterr = ::std::io::Error::last_os_error();
        if tmp == -1 {
            return Err(lasterr);
        }
        tmp.into()
    };

    if let Err(e) = enable_sock_opt(fd, ::libc::SO_REUSEPORT, 1) {
        #[allow(unsafe_code)]
        unsafe {
            ::libc::close(fd);
        }
        return Err(e);
    }
    if let Err(e) = enable_sock_opt(fd, ::libc::SO_REUSEADDR, 1) {
        #[allow(unsafe_code)]
        unsafe {
            ::libc::close(fd);
        }
        return Err(e);
    }

    // We will do path MTU discovery by ourselves,
    // always set the "don't fragment" bit
    let res = if addr.is_ipv6() {
        enable_sock_opt(fd, ::libc::IPV6_DONTFRAG, 1)
    } else {
        // FIXME: linux only
        enable_sock_opt(fd, ::libc::IP_MTU_DISCOVER, ::libc::IP_PMTUDISC_DO)
    };
    if let Err(e) = res {
        #[allow(unsafe_code)]
        unsafe {
            ::libc::close(fd);
        }
        return Err(e);
    }
    // manually convert rust SockAddr to C sockaddr
    #[allow(unsafe_code, trivial_casts, trivial_numeric_casts)]
    {
        let bind_ret = match addr {
            SocketAddr::V4(s4) => {
                let ip4: u32 = (*s4.ip()).into();
                let bind_addr = ::libc::sockaddr_in {
                    sin_family: ::libc::AF_INET as u16,
                    sin_port: s4.port().to_be(),
                    sin_addr: ::libc::in_addr { s_addr: ip4 },
                    sin_zero: [0; 8],
                };
                unsafe {
                    let c_addr =
                        &bind_addr as *const _ as *const ::libc::sockaddr;
                    ::libc::bind(fd, c_addr, 16)
                }
            }
            SocketAddr::V6(s6) => {
                let ip6: [u8; 16] = (*s6.ip()).octets();
                let bind_addr = ::libc::sockaddr_in6 {
                    sin6_family: ::libc::AF_INET6 as u16,
                    sin6_port: s6.port().to_be(),
                    sin6_flowinfo: 0,
                    sin6_addr: ::libc::in6_addr { s6_addr: ip6 },
                    sin6_scope_id: 0,
                };
                unsafe {
                    let c_addr =
                        &bind_addr as *const _ as *const ::libc::sockaddr;
                    ::libc::bind(fd, c_addr, 24)
                }
            }
        };
        let lasterr = ::std::io::Error::last_os_error();
        if bind_ret != 0 {
            unsafe {
                ::libc::close(fd);
            }
            return Err(lasterr);
        }
    }

    use ::std::os::fd::FromRawFd;
    #[allow(unsafe_code)]
    let std_sock = unsafe { ::std::net::UdpSocket::from_raw_fd(fd) };
    std_sock.set_nonblocking(true)?;
    ::tracing::debug!("Listening udp sock: {}", std_sock.local_addr().unwrap());

    Ok(UdpSocket::from_std(std_sock)?)
}
