use crate::{
    auth,
    connection::{handshake::*, ID},
    enc::{self, asym::KeyID},
};

#[test]
fn test_simple_encrypt_decrypt() {
    let rand = enc::Random::new();
    let cipher = enc::sym::CipherKind::XChaCha20Poly1305;
    let secret = enc::Secret::new_rand(&rand);
    let secret2 = secret.clone();

    let cipher_send = enc::sym::CipherSend::new(cipher, secret, &rand);
    let cipher_recv = enc::sym::CipherRecv::new(cipher, secret2);

    let mut data = Vec::new();
    let tot_len = cipher_recv.nonce_len().0 + 1234 + cipher_recv.tag_len().0;
    data.resize(tot_len, 0);
    rand.fill(&mut data);
    data[..enc::sym::Nonce::len()].copy_from_slice(&[0; 24]);
    let last = data.len() - cipher_recv.tag_len().0;
    data[last..].copy_from_slice(&[0; 16]);
    let orig = data.clone();
    let raw_aad: [u8; 8] = [0, 1, 2, 3, 4, 5, 6, 7];
    let aad = enc::sym::AAD(&raw_aad[..]);
    let aad2 = enc::sym::AAD(&raw_aad[..]);
    if cipher_send.encrypt(aad, &mut data).is_err() {
        assert!(false, "Encrypt failed");
    }
    if cipher_recv.decrypt(aad2, &mut data).is_err() {
        assert!(false, "Decrypt failed");
    }
    data[..enc::sym::Nonce::len()].copy_from_slice(&[0; 24]);
    let last = data.len() - cipher_recv.tag_len().0;
    data[last..].copy_from_slice(&[0; 16]);
    assert!(orig == data, "DIFFERENT!\n{:?}\n{:?}\n", orig, data);
}

#[test]
fn test_encrypt_decrypt() {
    let rand = enc::Random::new();
    let cipher = enc::sym::CipherKind::XChaCha20Poly1305;
    let secret = enc::Secret::new_rand(&rand);
    let secret2 = secret.clone();

    let cipher_send = enc::sym::CipherSend::new(cipher, secret, &rand);
    let cipher_recv = enc::sym::CipherRecv::new(cipher, secret2);
    let nonce_len = cipher_recv.nonce_len();
    let tag_len = cipher_recv.tag_len();

    let service_key = enc::Secret::new_rand(&rand);

    let data = dirsync::RespInner::ClearText(dirsync::RespData {
        client_nonce: dirsync::Nonce::new(&rand),
        id: ID::ID(::core::num::NonZeroU64::new(424242).unwrap()),
        service_connection_id: ID::ID(
            ::core::num::NonZeroU64::new(434343).unwrap(),
        ),
        service_key,
    });

    let resp = dirsync::Resp {
        client_key_id: KeyID(4444),
        data,
    };
    let encrypt_from = resp.encrypted_offset();
    let encrypt_to = encrypt_from + resp.encrypted_length(nonce_len, tag_len);

    let h_resp =
        Handshake::new(HandshakeData::DirSync(dirsync::DirSync::Resp(resp)));

    let mut bytes = Vec::<u8>::with_capacity(
        h_resp.len(cipher.nonce_len(), cipher.tag_len()),
    );
    bytes.resize(h_resp.len(cipher.nonce_len(), cipher.tag_len()), 0);
    h_resp.serialize(cipher.nonce_len(), cipher.tag_len(), &mut bytes);

    let raw_aad: [u8; 7] = [0, 1, 2, 3, 4, 5, 6];
    let aad = enc::sym::AAD(&raw_aad[..]);
    let aad2 = enc::sym::AAD(&raw_aad[..]);

    let pre_encrypt = bytes.clone();
    // encrypt
    if cipher_send
        .encrypt(aad, &mut bytes[encrypt_from..encrypt_to])
        .is_err()
    {
        assert!(false, "Encrypt failed");
    }
    if cipher_recv
        .decrypt(aad2, &mut bytes[encrypt_from..encrypt_to])
        .is_err()
    {
        assert!(false, "Decrypt failed");
    }
    // make sure Nonce and Tag are 0
    bytes[encrypt_from..(encrypt_from + nonce_len.0)].copy_from_slice(&[0; 24]);
    let tag_from = encrypt_to - tag_len.0;
    bytes[tag_from..(tag_from + tag_len.0)].copy_from_slice(&[0; 16]);
    assert!(
        pre_encrypt == bytes,
        "{}|{}=\n{:?}\n{:?}",
        encrypt_from,
        encrypt_to,
        pre_encrypt,
        bytes
    );

    // decrypt

    let mut deserialized = match Handshake::deserialize(&bytes) {
        Ok(deserialized) => deserialized,
        Err(e) => {
            assert!(false, "{}", e.to_string());
            return;
        }
    };
    // reparse
    if let HandshakeData::DirSync(dirsync::DirSync::Resp(r_a)) =
        &mut deserialized.data
    {
        let enc_start = r_a.encrypted_offset() + cipher.nonce_len().0;
        if let Err(e) = r_a.data.deserialize_as_cleartext(
            &bytes[enc_start..(bytes.len() - cipher.tag_len().0)],
        ) {
            assert!(false, "DirSync Resp Inner serialize: {}", e.to_string());
        }
    };

    assert!(
        deserialized == h_resp,
        "DirSync Resp (de)serialization not working",
    );
}
