//! Worker thread implementation
use crate::{
    auth::{self, Domain, ServiceID, Token, TokenChecker, UserID},
    config::Config,
    connection::{
        self,
        handshake::{
            self,
            dirsync::{self, DirSync},
            tracker::{HandshakeAction, HandshakeTracker},
            Handshake, HandshakeData,
        },
        socket::{UdpClient, UdpServer},
        ConnList, Connection, IDSend, Packet,
    },
    dnssec,
    enc::{
        asym::{self, KeyID, PrivKey, PubKey},
        hkdf::{self, Hkdf, HkdfKind},
        sym, Random, Secret,
    },
    inner::ThreadTracker,
};
use ::std::{sync::Arc, vec::Vec};
/// This worker must be cpu-pinned
use ::tokio::{
    net::UdpSocket,
    sync::{mpsc, oneshot, Mutex},
};

/// Track a raw Udp packet
pub(crate) struct RawUdp {
    pub src: UdpClient,
    pub dst: UdpServer,
    pub data: Vec<u8>,
    pub packet: Packet,
}

pub(crate) struct ConnectInfo {
    pub answer: oneshot::Sender<handshake::tracker::ConnectAnswer>,
    pub resolved: dnssec::Record,
    pub service_id: ServiceID,
    pub domain: Domain,
    // TODO: UserID, Token information
}

pub(crate) enum Work {
    /// ask the thread to report to the main thread the total number of
    /// connections present
    CountConnections(oneshot::Sender<usize>),
    Connect(ConnectInfo),
    DropHandshake(KeyID),
    Recv(RawUdp),
}

/// Actual worker implementation.
#[allow(missing_debug_implementations)]
pub struct Worker {
    cfg: Config,
    thread_id: ThreadTracker,
    // PERF: rand uses syscalls. how to do that async?
    rand: Random,
    stop_working: crate::StopWorkingRecvCh,
    token_check: Option<Arc<Mutex<TokenChecker>>>,
    sockets: Vec<Arc<UdpSocket>>,
    queue: ::async_channel::Receiver<Work>,
    queue_timeouts_recv: mpsc::UnboundedReceiver<Work>,
    queue_timeouts_send: mpsc::UnboundedSender<Work>,
    thread_channels: Vec<::async_channel::Sender<Work>>,
    connections: ConnList,
    handshakes: HandshakeTracker,
}

#[allow(unsafe_code)]
unsafe impl Send for Worker {}

impl Worker {
    pub(crate) async fn new(
        mut cfg: Config,
        thread_id: ThreadTracker,
        stop_working: crate::StopWorkingRecvCh,
        token_check: Option<Arc<Mutex<TokenChecker>>>,
        sockets: Vec<Arc<UdpSocket>>,
        queue: ::async_channel::Receiver<Work>,
    ) -> ::std::io::Result<Self> {
        let (queue_timeouts_send, queue_timeouts_recv) =
            mpsc::unbounded_channel();
        let mut handshakes = HandshakeTracker::new(
            thread_id,
            cfg.ciphers.clone(),
            cfg.key_exchanges.clone(),
        );
        let mut server_keys = Vec::new();
        // make sure the keys are no longer in the config
        ::core::mem::swap(&mut server_keys, &mut cfg.server_keys);
        for k in server_keys.into_iter() {
            if handshakes.add_server_key(k.id, k.priv_key).is_err() {
                return Err(::std::io::Error::new(
                    ::std::io::ErrorKind::AlreadyExists,
                    "You can't use the same KeyID for multiple keys",
                ));
            }
        }
        for srv in cfg.servers.iter() {
            if handshakes.add_server_domain(&srv.fqdn, &srv.keys).is_err() {
                return Err(::std::io::Error::new(
                    ::std::io::ErrorKind::NotFound,
                    "Specified a KeyID that we don't have",
                ));
            }
        }

        Ok(Self {
            cfg,
            thread_id,
            rand: Random::new(),
            stop_working,
            token_check,
            sockets,
            queue,
            queue_timeouts_recv,
            queue_timeouts_send,
            thread_channels: Vec::new(),
            connections: ConnList::new(thread_id),
            handshakes,
        })
    }
    /// Continuously loop and process work as needed
    pub async fn work_loop(&mut self) {
        'mainloop: loop {
            let work = ::tokio::select! {
                tell_stopped = self.stop_working.recv() => {
                    if let Ok(stop_ch) = tell_stopped {
                        let _ = stop_ch.send(
                             crate::StopWorking::WorkerStopped).await;
                    }
                    break;
                }
                maybe_timeout = self.queue.recv() => {
                    match maybe_timeout {
                        Ok(work) => work,
                        Err(_) => break,
                    }
                }
                maybe_work = self.queue.recv() => {
                    match maybe_work {
                        Ok(work) => work,
                        Err(_) => break,
                    }
                }
            };
            match work {
                Work::CountConnections(sender) => {
                    let conn_num = self.connections.len();
                    let _ = sender.send(conn_num);
                }
                Work::Connect(conn_info) => {
                    // PERF: geolocation

                    // Find the first destination with:
                    // * UDP port
                    // * a coherent pubkey/key exchange.
                    let destination =
                        conn_info.resolved.addresses.iter().find_map(|addr| {
                            if addr.port.is_none()
                                || addr
                                    .handshake_ids
                                    .iter()
                                    .find(|h_srv| {
                                        self.cfg.handshakes.contains(h_srv)
                                    })
                                    .is_none()
                            {
                                // skip servers with no corresponding
                                // handshake types or no udp port
                                return None;
                            }

                            // make sure this server has a public key
                            // that supports one of the key exchanges that
                            // *we* support
                            for idx in addr.public_key_idx.iter() {
                                // for each key,
                                // get all the possible key exchanges
                                let key_supported_k_x =
                                    conn_info.resolved.public_keys
                                        [idx.0 as usize]
                                        .1
                                        .kind()
                                        .key_exchanges();
                                // consider only the key exchanges allowed
                                // in the dnssec record
                                let filtered_key_exchanges: Vec<
                                    asym::KeyExchangeKind,
                                > = key_supported_k_x
                                    .into_iter()
                                    .filter(|k_x| {
                                        conn_info
                                            .resolved
                                            .key_exchanges
                                            .contains(k_x)
                                    })
                                    .copied()
                                    .collect();

                                // finally make sure that we support those
                                match self.cfg.key_exchanges.iter().find(|x| {
                                    filtered_key_exchanges.contains(x)
                                }) {
                                    Some(exchange) => {
                                        return Some((
                                            addr,
                                            conn_info.resolved.public_keys
                                                [idx.0 as usize],
                                            exchange.clone(),
                                        ))
                                    }
                                    None => return None,
                                }
                            }
                            return None;
                        });
                    let (addr, key, exchange) = match destination {
                        Some((addr, key, exchange)) => (addr, key, exchange),
                        None => {
                            let _ =
                                conn_info
                                    .answer
                                    .send(Err(crate::Error::Resolution(
                                    "No selectable address and key combination"
                                        .to_owned(),
                                )));
                            continue 'mainloop;
                        }
                    };
                    let hkdf_selected = match hkdf::client_select_hkdf(
                        &self.cfg,
                        &conn_info.resolved.hkdfs,
                    ) {
                        Some(hkdf_selected) => hkdf_selected,
                        None => {
                            let _ = conn_info.answer.send(Err(
                                handshake::Error::Negotiation.into(),
                            ));
                            continue 'mainloop;
                        }
                    };
                    let cipher_selected = match sym::client_select_cipher(
                        &self.cfg,
                        &conn_info.resolved.ciphers,
                    ) {
                        Some(cipher_selected) => cipher_selected,
                        None => {
                            let _ = conn_info.answer.send(Err(
                                handshake::Error::Negotiation.into(),
                            ));
                            continue 'mainloop;
                        }
                    };

                    let (priv_key, pub_key) =
                        match exchange.new_keypair(&self.rand) {
                            Ok(pair) => pair,
                            Err(_) => {
                                ::tracing::error!("Failed to generate keys");
                                let _ = conn_info.answer.send(Err(
                                    handshake::Error::KeyGeneration.into(),
                                ));
                                continue 'mainloop;
                            }
                        };
                    let hkdf;

                    if let PubKey::Exchange(srv_pub) = key.1 {
                        let secret =
                            match priv_key.key_exchange(exchange, srv_pub) {
                                Ok(secret) => secret,
                                Err(_) => {
                                    ::tracing::warn!(
                                        "Could not run the key exchange"
                                    );
                                    let _ = conn_info.answer.send(Err(
                                        handshake::Error::Negotiation.into(),
                                    ));
                                    continue 'mainloop;
                                }
                            };
                        hkdf = Hkdf::new(hkdf_selected, b"fenrir", secret);
                    } else {
                        // crate::dnssec already verifies that the keys
                        // listed in dnssec::Record.addresses.public_key_idx
                        // are PubKey::Exchange
                        unreachable!()
                    }
                    let mut conn = Connection::new(
                        hkdf,
                        cipher_selected,
                        connection::Role::Client,
                        &self.rand,
                    );

                    let auth_recv_id = self.connections.reserve_first();
                    let service_conn_id = self.connections.reserve_first();
                    conn.id_recv = auth_recv_id;
                    let (client_key_id, hshake) = match self
                        .handshakes
                        .add_client(
                            PrivKey::Exchange(priv_key),
                            PubKey::Exchange(pub_key),
                            conn_info.service_id,
                            service_conn_id,
                            conn,
                            conn_info.answer,
                            key.0,
                        ) {
                        Ok((client_key_id, hshake)) => (client_key_id, hshake),
                        Err(answer) => {
                            ::tracing::warn!("Too many client handshakes");
                            let _ = answer.send(Err(
                                handshake::Error::TooManyClientHandshakes
                                    .into(),
                            ));
                            continue 'mainloop;
                        }
                    };

                    // build request
                    let auth_info = dirsync::AuthInfo {
                        user: UserID::new_anonymous(),
                        token: Token::new_anonymous(&self.rand),
                        service_id: conn_info.service_id,
                        domain: conn_info.domain,
                    };
                    let req_data = dirsync::ReqData {
                        nonce: dirsync::Nonce::new(&self.rand),
                        client_key_id,
                        id: auth_recv_id.0, //FIXME: is zero
                        auth: auth_info,
                    };
                    let req = dirsync::Req {
                        key_id: key.0,
                        exchange,
                        hkdf: hkdf_selected,
                        cipher: cipher_selected,
                        exchange_key: pub_key,
                        data: dirsync::ReqInner::ClearText(req_data),
                    };
                    let encrypt_start = ID::len() + req.encrypted_offset();
                    let encrypt_end = encrypt_start
                        + req.encrypted_length(
                            cipher_selected.nonce_len(),
                            cipher_selected.tag_len(),
                        );
                    let h_req = Handshake::new(HandshakeData::DirSync(
                        DirSync::Req(req),
                    ));
                    use connection::{PacketData, ID};
                    let packet = Packet {
                        id: ID::Handshake,
                        data: PacketData::Handshake(h_req),
                    };

                    let tot_len = packet.len(
                        cipher_selected.nonce_len(),
                        cipher_selected.tag_len(),
                    );
                    let mut raw = Vec::<u8>::with_capacity(tot_len);
                    raw.resize(tot_len, 0);
                    packet.serialize(
                        cipher_selected.nonce_len(),
                        cipher_selected.tag_len(),
                        &mut raw[..],
                    );
                    // encrypt
                    if let Err(e) = hshake.connection.cipher_send.encrypt(
                        sym::AAD(&[]),
                        &mut raw[encrypt_start..encrypt_end],
                    ) {
                        ::tracing::error!("Can't encrypt DirSync Request");
                        if let Some(client) =
                            self.handshakes.remove_client(client_key_id)
                        {
                            let _ = client.answer.send(Err(e.into()));
                        };
                        continue 'mainloop;
                    }

                    // send always from the first socket
                    // FIXME: select based on routing table
                    let sender = self.sockets[0].local_addr().unwrap();
                    let dest = UdpClient(addr.as_sockaddr().unwrap());

                    // start the timeout right before sending the packet
                    hshake.timeout = Some(::tokio::task::spawn_local(
                        Self::handshake_timeout(
                            self.queue_timeouts_send.clone(),
                            client_key_id,
                        ),
                    ));

                    // send packet
                    self.send_packet(raw, dest, UdpServer(sender)).await;

                    continue 'mainloop;
                }
                Work::DropHandshake(key_id) => {
                    if let Some(connections) =
                        self.handshakes.timeout_client(key_id)
                    {
                        for conn_id in connections.into_iter() {
                            self.connections.remove(conn_id);
                        }
                    };
                }
                Work::Recv(pkt) => {
                    self.recv(pkt).await;
                }
            }
        }
    }
    async fn handshake_timeout(
        timeout_queue: mpsc::UnboundedSender<Work>,
        key_id: KeyID,
    ) {
        ::tokio::time::sleep(::std::time::Duration::from_secs(10)).await;
        let _ = timeout_queue.send(Work::DropHandshake(key_id));
    }
    /// Read and do stuff with the raw udp packet
    async fn recv(&mut self, mut udp: RawUdp) {
        if udp.packet.id.is_handshake() {
            let handshake = match Handshake::deserialize(
                &udp.data[connection::ID::len()..],
            ) {
                Ok(handshake) => handshake,
                Err(e) => {
                    ::tracing::debug!("Handshake parsing: {}", e);
                    return;
                }
            };
            let action = match self.handshakes.recv_handshake(
                handshake,
                &mut udp.data[connection::ID::len()..],
            ) {
                Ok(action) => action,
                Err(err) => {
                    ::tracing::debug!("Handshake recv error {}", err);
                    return;
                }
            };
            match action {
                HandshakeAction::AuthNeeded(authinfo) => {
                    let req;
                    if let HandshakeData::DirSync(DirSync::Req(r)) =
                        authinfo.handshake.data
                    {
                        req = r;
                    } else {
                        ::tracing::error!("AuthInfo on non DS::Req");
                        return;
                    }
                    use dirsync::ReqInner;
                    let req_data = match req.data {
                        ReqInner::ClearText(req_data) => req_data,
                        _ => {
                            ::tracing::error!("AuthNeeded: expected ClearText");
                            assert!(false, "AuthNeeded: unreachable");
                            return;
                        }
                    };
                    // FIXME: This part can take a while,
                    // we should just spawn it probably
                    let maybe_auth_check = {
                        match &self.token_check {
                            None => {
                                if req_data.auth.user == auth::USERID_ANONYMOUS
                                {
                                    Ok(true)
                                } else {
                                    Ok(false)
                                }
                            }
                            Some(token_check) => {
                                let tk_check = token_check.lock().await;
                                tk_check(
                                    req_data.auth.user,
                                    req_data.auth.token,
                                    req_data.auth.service_id,
                                    req_data.auth.domain,
                                )
                                .await
                            }
                        }
                    };
                    let is_authenticated = match maybe_auth_check {
                        Ok(is_authenticated) => is_authenticated,
                        Err(_) => {
                            ::tracing::error!("error in token auth");
                            // TODO: retry?
                            return;
                        }
                    };
                    if !is_authenticated {
                        ::tracing::warn!(
                            "Wrong authentication for user {:?}",
                            req_data.auth.user
                        );
                        // TODO: error response
                        return;
                    }
                    // Client has correctly authenticated
                    // TODO: contact the service, get the key and
                    // connection ID
                    let srv_conn_id = ID::new_rand(&self.rand);
                    let srv_secret = Secret::new_rand(&self.rand);
                    let head_len = req.cipher.nonce_len();
                    let tag_len = req.cipher.tag_len();

                    let mut auth_conn = Connection::new(
                        authinfo.hkdf,
                        req.cipher,
                        connection::Role::Server,
                        &self.rand,
                    );
                    auth_conn.id_send = IDSend(req_data.id);
                    // track connection
                    let auth_id_recv = self.connections.reserve_first();
                    auth_conn.id_recv = auth_id_recv;

                    let resp_data = dirsync::RespData {
                        client_nonce: req_data.nonce,
                        id: auth_conn.id_recv.0,
                        service_connection_id: srv_conn_id,
                        service_key: srv_secret,
                    };
                    use crate::enc::sym::AAD;
                    // no aad for now
                    let aad = AAD(&mut []);

                    use dirsync::RespInner;
                    let resp = dirsync::Resp {
                        client_key_id: req_data.client_key_id,
                        data: RespInner::ClearText(resp_data),
                    };
                    let encrypt_from = ID::len() + resp.encrypted_offset();
                    let encrypt_until =
                        encrypt_from + resp.encrypted_length(head_len, tag_len);
                    let resp_handshake = Handshake::new(
                        HandshakeData::DirSync(DirSync::Resp(resp)),
                    );
                    use connection::{PacketData, ID};
                    let packet = Packet {
                        id: ID::new_handshake(),
                        data: PacketData::Handshake(resp_handshake),
                    };
                    let tot_len = packet.len(head_len, tag_len);
                    let mut raw_out = Vec::<u8>::with_capacity(tot_len);
                    raw_out.resize(tot_len, 0);
                    packet.serialize(head_len, tag_len, &mut raw_out);

                    if let Err(e) = auth_conn
                        .cipher_send
                        .encrypt(aad, &mut raw_out[encrypt_from..encrypt_until])
                    {
                        ::tracing::error!("can't encrypt: {:?}", e);
                        return;
                    }
                    self.send_packet(raw_out, udp.src, udp.dst).await;
                }
                HandshakeAction::ClientConnect(cci) => {
                    let ds_resp;
                    if let HandshakeData::DirSync(DirSync::Resp(resp)) =
                        cci.handshake.data
                    {
                        ds_resp = resp;
                    } else {
                        ::tracing::error!("ClientConnect on non DS::Resp");
                        return;
                    }
                    // track connection
                    let resp_data;
                    if let dirsync::RespInner::ClearText(r_data) = ds_resp.data
                    {
                        resp_data = r_data;
                    } else {
                        ::tracing::error!(
                            "ClientConnect on non DS::Resp::ClearText"
                        );
                        unreachable!();
                    }
                    let auth_srv_conn = IDSend(resp_data.id);
                    let mut conn = cci.connection;
                    conn.id_send = auth_srv_conn;
                    let id_recv = conn.id_recv;
                    let cipher = conn.cipher_recv.kind();
                    // track the connection to the authentication server
                    if self.connections.track(conn.into()).is_err() {
                        ::tracing::error!("Could not track new connection");
                        self.connections.remove(id_recv);
                        let _ = cci.answer.send(Err(
                            handshake::Error::InternalTracking.into(),
                        ));
                        return;
                    }
                    if cci.service_id != auth::SERVICEID_AUTH {
                        // create and track the connection to the service
                        // SECURITY: xor with secrets
                        //FIXME: the Secret should be XORed with the client
                        // stored secret (if any)
                        let hkdf = Hkdf::new(
                            HkdfKind::Sha3,
                            cci.service_id.as_bytes(),
                            resp_data.service_key,
                        );
                        let mut service_connection = Connection::new(
                            hkdf,
                            cipher,
                            connection::Role::Client,
                            &self.rand,
                        );
                        service_connection.id_recv = cci.service_connection_id;
                        service_connection.id_send =
                            IDSend(resp_data.service_connection_id);
                        let _ =
                            self.connections.track(service_connection.into());
                    }
                    let _ =
                        cci.answer.send(Ok((cci.srv_key_id, auth_srv_conn)));
                }
                HandshakeAction::Nothing => {}
            };
        }
    }
    async fn send_packet(
        &self,
        data: Vec<u8>,
        client: UdpClient,
        server: UdpServer,
    ) {
        let src_sock = match self
            .sockets
            .iter()
            .find(|&s| s.local_addr().unwrap() == server.0)
        {
            Some(src_sock) => src_sock,
            None => {
                ::tracing::error!(
                    "Can't send packet: Server changed listening ip{}!",
                    server.0
                );
                return;
            }
        };
        let _res = src_sock.send_to(&data, client.0).await;
    }
}
